module Hence
  module Emulator
    extend self

    def emulate(data : Data) : Data
      until data.reg_pc == 0xffff_u16
        data.reg_opc = (data.get_memory(data.reg_pc) & 0xff_u8).to_u8
        data.reg_pc &+= 1
        data.reg_arg = data.get_memory(data.reg_pc) << 8
        data.reg_pc &+= 1
        data.reg_arg = data.reg_arg | (data.get_memory(data.reg_pc) & 0xff_u8)
        data.reg_pc &+= 1

        base_mcc = data.reg_opc &* 4
        data.reg_mccs = data.firmware[base_mcc].to_u16 << 8
        data.reg_mccs = data.reg_mccs | data.firmware[base_mcc &+ 1]
        data.reg_mcce = data.firmware[base_mcc &+ 2].to_u16 << 8
        data.reg_mcce = data.reg_mcce | data.firmware[base_mcc &+ 3]

        until data.reg_mccs &+ data.reg_mccpc >= data.reg_mcce
          base_mccpc = data.reg_mccs &+ data.reg_mccpc
          data.reg_mcc = data.firmware[base_mccpc]
          data.reg_mccarg = data.firmware[base_mccpc &+ 1].to_u16 << 8
          data.reg_mccarg = data.reg_mccarg | data.firmware[base_mccpc &+ 2]
          data.reg_mccpc &+= 3

          data.microcodes[data.reg_mcc].run(data)
        end

        data.reg_mccpc = 0_u16
        data.reg_tmp = 0_u16
        data.reg_a = 0_u16
        data.reg_b = 0_u16
        data.reg_c = 0_u16
        data.reg_d = 0_u16
      end

      data
    end

    class Data
      property firmware
      property microcodes
      property reg_pc
      property reg_opc
      property reg_arg
      property reg_mccpc
      property reg_mccs
      property reg_mcce
      property reg_mcc
      property reg_mccarg
      property reg_tmp
      property reg_sp
      property reg_a
      property reg_b
      property reg_c
      property reg_d
      property reg_inp
      property reg_out

      def initialize(
        @program : Bytes,
        @firmware : Bytes,
        @microcodes = Microcode::MICROCODES,
        @reg_pc = 0_u16,
        @reg_opc = 0_u8,
        @reg_arg = 0_u16,
        @reg_mccpc = 0_u16,
        @reg_mccs = 0_u16,
        @reg_mcce = 0_u16,
        @reg_mcc = 0_u8,
        @reg_mccarg = 0_u16,
        @reg_tmp = 0_u16,
        @reg_sp = 0_u16,
        @reg_a = 0_u16,
        @reg_b = 0_u16,
        @reg_c = 0_u16,
        @reg_d = 0_u16,
        @reg_inp = 0_u16,
        @reg_out = 0_u16,
        @stack = Slice(UInt16).new({{ 8 * 1024 }}),
        @memory = Slice(UInt16).new({{ 16 * 1024 }})
      )
      end

      def get_register(index : UInt8) : UInt16
        case index
        when 0x0_u8
          @reg_pc
        when 0x1_u8
          @reg_opc.to_u16
        when 0x2_u8
          @reg_arg
        when 0x3_u8
          @reg_mccpc
        when 0x4_u8
          @reg_mccs
        when 0x5_u8
          @reg_mcce
        when 0x6_u8
          @reg_mcc.to_u16
        when 0x7_u8
          @reg_mccarg
        when 0x8_u8
          @reg_tmp
        when 0x9_u8
          @reg_sp
        when 0xa_u8
          @reg_a
        when 0xb_u8
          @reg_b
        when 0xc_u8
          @reg_c
        when 0xd_u8
          @reg_d
        when 0xe_u8
          @reg_inp
        when 0xf_u8
          @reg_out
        else
          raise "Invalid register index: #{index}"
        end
      end

      def set_register(index : UInt8, value : UInt16) : UInt16
        case index
        when 0x0_u8
          @reg_pc = value
        when 0x1_u8
          @reg_opc = value
        when 0x2_u8
          @reg_arg = value
        when 0x3_u8
          @reg_mccpc = value
        when 0x4_u8
          @reg_mccs = value
        when 0x5_u8
          @reg_mcce = value
        when 0x6_u8
          @reg_mcc = value
        when 0x7_u8
          @reg_mccarg = value
        when 0x8_u8
          @reg_tmp = value
        when 0x9_u8
          @reg_sp = value
        when 0xa_u8
          @reg_a = value
        when 0xb_u8
          @reg_b = value
        when 0xc_u8
          @reg_c = value
        when 0xd_u8
          @reg_d = value
        when 0xe_u8
          @reg_inp = value
        when 0xf_u8
          @reg_out = value
        else
          raise "Invalid register index: #{index}"
        end
      end

      def alu(op : UInt8) : UInt16
        @reg_tmp =
          case op
          when 0x00 # not
            ~@reg_a
          when 0x01 # and
            @reg_a & @reg_b
          when 0x02 # nand
            ~(@reg_a & @reg_b)
          when 0x03 # or
            @reg_a | @reg_b
          when 0x04 # nor
            ~(@reg_a | @reg_b)
          when 0x05 # xor
            @reg_a ^ @reg_b
          when 0x06 # xnor
            ~(@reg_a ^ @reg_b)
          when 0x07 # shl
            @reg_a << @reg_b
          when 0x08 # shr
            @reg_a >> @reg_b
          when 0x09 # add
            @reg_a &+ @reg_b
          when 0x0a # sub
            @reg_a &- @reg_b
          when 0x0b # mul
            @reg_a &* @reg_b
          when 0x0c # div
            @reg_a // @reg_b
          when 0x0d # mod
            @reg_a % @reg_b
          when 0x0e # cmp
            if @reg_a < @reg_b
              1_u16
            elsif @reg_a > @reg_b
              2_u16
            else
              0_u16
            end
          when 0x0f # eq
            @reg_a == @reg_b ? 1_u16 : 0_u16
          when 0x10 # neq
            @reg_a != @reg_b ? 1_u16 : 0_u16
          when 0x11 # lt
            @reg_a < @reg_b ? 1_u16 : 0_u16
          when 0x12 # gt
            @reg_a > @reg_b ? 1_u16 : 0_u16
          when 0x13 # leq
            @reg_a <= @reg_b ? 1_u16 : 0_u16
          when 0x14 # geq
            @reg_a >= @reg_b ? 1_u16 : 0_u16
          when 0x15 # bol
            @reg_a == 1 ? 1_u16 : 0_u16
          when 0x16 # neg
            @reg_a == 1 ? 0_u16 : 1_u16
          else
            raise "Invalid ALU operation: 0x#{op.to_s(16).rjust(2, '0')}"
          end
      end

      def get_memory(address : UInt16) : UInt16
        if address < {{ 32 * 1024 }}
          @program[address].to_u16
        elsif address < {{ 40 * 1024 }}
          @stack[address - {{ 32 * 1024 }}]
        elsif address < {{ 56 * 1024 }}
          @memory[address - {{ 40 * 1024 }}]
        else
          0_u16
        end
      end

      def set_memory(address : UInt16, value : UInt16) : UInt16
        if address < {{ 32 * 1024 }}
          0_u16
        elsif address < {{ 40 * 1024 }}
          @stack[address - {{ 32 * 1024 }}] = value

          value
        elsif address < {{ 56 * 1024 }}
          @memory[address - {{ 40 * 1024 }}] = value
        else
          0_u16
        end
      end
    end
  end
end
